import json
import os
import sqlite3
import pprint
import psycopg2
import psycopg2.extras
import traceback
import sys
import pytz
import logging
from flask import g

from fk.db.DatabaseConnection import DatabaseConnection


logger = logging.getLogger(__name__)


class Database:
    def __init__(self, config):
        self.config = config
        self.dbc = DatabaseConnection.get_connection(self.config)
        assert self.dbc.is_ok()
        self.create_tables()

    def create_tables(self):

        # Create a table to keep track of crons
        self.dbc.query_none(
            """
			create table if not exists "clockwork_crons" (
				id serial primary key,
				name varchar(31),
				spec text,
				command varchar(255),
				arguments text, 
				enabled boolean not null default true,
				run_count bigint not null default 0, 
				last_run_at timestamptz not null default now(),
				created_at timestamptz not null default now(),
				updated_at timestamptz not null default now()
			);
			comment on column clockwork_crons.id is 'Unique internal id for this cron';
			comment on column clockwork_crons.name is 'The cron''s name.';
			comment on column clockwork_crons.spec is 'The cron''s time specification.';
			comment on column clockwork_crons.command is 'The cron''s command to execute.';
			comment on column clockwork_crons.arguments is 'The cron''s command arguments.';
			comment on column clockwork_crons.last_run_at is 'When the cron was last run';
			comment on column clockwork_crons.created_at is 'When the cron was first created';
			comment on column clockwork_crons.updated_at is 'When cron item was last updated';
			"""
        )

    # Insert a new cron into
    def upsert_cron(self, cron):
        return self.dbc.query_one(
            """
				insert into clockwork_crons
					(
					name,
					spec,
					command,
					arguments, 
					enabled,
					run_count,
					last_run_at
					)
				values
					(
					%(name)s,
					%(spec)s,
					%(command)s,
					%(arguments)s,
					%(enabled)s,
					%(run_count)s,
					%(last_run_at)s
					)
				on conflict(id)
				do update
					set
					name=%(name)s,
					spec=%(spec)s,
					command=%(command)s,
					arguments=%(arguments)s,
					enabled=%(enabled)s,
					run_count=%(run_count)s,
					last_run_at=%(last_run_at)s,
					updated_at=now()
				returning id
				;
				""",
            cron,
        )

    # Register a new cron execution
    def log_cron_run(self, id):
        return self.dbc.query_one(
            """
				update
					clockwork_crons
				set
					run_count = run_count + 1,
					last_run_at = now(),
					updated_at = now()
				where
					id = %(id)s
				returning
					updated_at
				;
				""",
            {"id": (id,)},
        )

    # Update a cron with a new enabled state
    def set_cron_enabled(self, id, enabled=True):
        return self.dbc.query_one(
            """
                update
                    clockwork_crons
                set
                    enabled = %(enabled)s
                    updated_at = now()
                where
                    id = %(id)s
                returning
                    updated_at
                ;
                """,
            {"id": (id,), "enabled": (enabled,)},
        )

        # Delete cron with given id

    def delete_cron_with_id(self, cron_id):
        self.dbc.query_none(
            """
				delete from clockwork_crons
				where id=%(cron_id)s,
				;
				""",
            {"cron_id": (cron_id,)},
        )

    # Clear out the crons table
    def delete_all(self):
        return self.dbc.query_none(
            """
				delete from clockwork_crons
				;
				"""
        )

    # Get crons from cron list fitting the optional filter
    def get_crons(self, id=None, name=None, spec=None, command=None, arguments=None, enabled=None, run_count=None, last_run_at=None, limit=1):
        return self.dbc.query_many(
            """
				select
					id,
					name,
					spec,
					command,
					arguments, 
					enabled,
					run_count,
					last_run_at,
					created_at,
					updated_at
				from
					clockwork_crons
				where
					true
				and
					(%(id)s is null or id = any(%(id)s))
				and
					(%(name)s is null or name = any(%(name)s))
				and
					(%(spec)s is null or spec = any(%(spec)s))
				and
					(%(command)s is null or command = any(%(command)s))
				and
					(%(arguments)s is null or arguments = any(%(arguments)s))
				and
					(%(enabled)s is null or enabled = any(%(enabled)s))
				and
					(%(run_count)s is null or run_count = any(%(run_count)s))
				and
					(%(last_run_at)s is null or last_run_at = any(%(last_run_at)s))
				order by
					last_run_at asc
				limit
					%(limit)s
				;
				""",
            {"id": (id,), "enabled": (enabled,), "name": (name,), "spec": (spec,), "command": (command,), "arguments": (arguments,), "run_count": (run_count,), "last_run_at": (last_run_at,), "limit": (limit,)},
        )

    # Get cron by id from cron list
    def get_cron_by_id(self, id):
        return self.dbc.query_one(
            """
				select
					id,
					name,
					spec,
					command,
					arguments, 
					enabled,
					run_count,
					last_run_at,
					created_at,
					updated_at
				from
					clockwork_crons
				where
					id = %(id)s
				order by
					updated_at asc
				limit
					1
				;
				""",
            {"id": (id,)},
        )

    def get_now(self):
        r = self.dbc.query_one(
            """
				select now()
			;"""
        )
        if not r:
            return None
        r = r.get("now")
        if not r:
            return None
        # logger.info(r)
        r = r.replace(tzinfo=pytz.UTC)
        return r


# What is this shit you ask? Gotta love python.... https://flask.palletsprojects.com/en/1.1.x/tutorial/database/
def get_flask_clockwork_db(config) -> Database:
    if "clockwork_database" not in g:
        g.clockwork_database = Database(config)
    return g.clockwork_database


def close_flask_clockwork_db(e=None):
    clockwork_database = g.pop("clockwork_database", None)

    if clockwork_database is not None:
        del clockwork_database



from fk.clockwork.db import Database
from pprint import pformat
import datetime
import fk.utils
import fk.utils.Watchdog
import logging
import pytz
import re
import schedule
import time
import traceback

logger = logging.getLogger(__name__)


def job():
    logger.info("Performing dummy job")


props = ["seconds", "minutes", "hours", "days", "weeks", "second", "minute", "hour", "day", "week", "monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday"]


def parse_schedule(spec, job=job, tag=None):
    # logger.info(f"Processing spec: {spec}")
    if str != type(spec):
        logger.error("Spec not string")
        return None
    spec = "." + spec.strip()
    if not spec:
        logger.error("Empty spec")
        return None
    parts = spec.split(".")
    if not parts:
        logger.error("Could not split spec")
        return None
    if parts[0] == "schedule":
        del parts[0]
    better = "." + ".".join(parts)
    regex = r"\.(?P<op>every|to|at|do|" + "|".join(props) + r')(?:\([\'"]?(?P<arg>job|[1-9]?[0-9]*|[1-9][0-9]*:[0-9][0-9]|:[0-9][0-9])[\'"]?\))?'
    # logger.info(f"REGEX IS: {regex}")
    matches = list(re.finditer(regex, better))
    # logger.info(f"Found total of {len(matches)} matches")
    if not matches:
        logger.error("No matches in spec")
        return None
    first = True
    base = None
    saw_job = False
    processed = 0
    for match in matches:
        groups = match.groupdict()
        # logger.info(groups)
        op = groups.get("op")
        arg = groups.get("arg")
        try:
            arg = int(arg)
        except:
            pass
        # logger.info(f"Match[{processed}]: {op}({arg}) ----")
        processed += 1
        try:
            if first:
                first = False
                if op != "every":
                    logger.error("Spec did not start with 'every'")
                    return None
                base = schedule.every(int(float(arg or 1)))
                if tag:
                    base.tag(tag)
                # logger.info("Created schedule")
                continue
            if saw_job:
                logger.error("Commands after job")
                return None
            if op == "every":
                logger.error("Invalid second call to every")
                return None
            if op in props:
                base = getattr(base, op)
                # logger.info(f"Added {op}")
                continue
            if op in ["to", "at"]:
                func = getattr(base, op)
                base = func(arg)
                # logger.info(f"Added {op}({arg})")
                continue
            if op == "do":
                if arg != "job":
                    logger.error(f"Invalid job {arg}")
                    return None
                base = base.do(job)
                # logger.info(f"Finalized")
                saw_job = True
                continue
        except Exception as e:
            logger.error(f'Failed to parse schedule with "{e}"')
            return None
    # Output: ['cats', 'dogs']
    # logger.info(f"Processed total of {processed} matches")
    return base
    # [x.group() for x in re.finditer( r'all (.*?) are', 'all cats are smarter than dogs, all dogs are dumber than cats')]
    # Output: ['all cats are', 'all dogs are']


class Processor:
    def __init__(self, config):
        self.config = config
        self.db = Database(self.config)
        self.task_filter = re.compile("^([0-9a-z\-_]+)$")
        self.last_status_time = None
        self.callables = {}
        self.last_time = self.db.get_now()
        self.universal_basetime = datetime.datetime(1970, 1, 1, 0, 0, 0, 0, pytz.UTC)

    def _execute_safely(self, entrypoint, item):
        # logger.info("SAFE EXECUTION STARTED!")
        # We cap runtime at 60 seconds
        timeout = 30
        failure = None
        result = None
        try:
            watchdog = fk.utils.Watchdog.Watchdog(timeout)
            # logger.info("££££ Entry")
            result, failure = entrypoint(item, self.config)
            # logger.info("££££ Exit")
        except fk.utils.Watchdog.Watchdog:
            logger.warning("Watchdag triggered exception")
            failure = f"execution timed out after {timeout} seconds"
        except Exception as e:
            logger.warning("")
            logger.warning("")
            logger.warning(f"###############################################")
            logger.warning(f"#    Batch item: {item}")
            logger.warning(f"# Failed with: {e}")
            failure = f"{e}"
            logger.warning(f"#          At:")
            traceback.print_exc()
            logger.warning(f"#######################################")
            logger.warning("")
            logger.warning("")
        watchdog.stop()
        # logger.info("SAFE EXECUTION FINISHED!")
        return result, failure

    def execute_one_cron(self):
        """
        Take ownership of one cron and make sure it is properly executed and updated with status underway
        """
        item = self.db.book_batch_item(self.db.PENDING, self.db.IN_PROGRESS)
        if item:
            logger.info("Processing item:")
            logger.info(pformat(item))
            id = item.get("id", None)
            type = item.get("type", None)
            updated_at = item.get("updated_at", None)
            if id and type and updated_at:
                entrypoint, failure = self.get_callable_for_type(type)
                if entrypoint and failure:
                    entrypoint = None
                    failure = f"{failure} AND ENTRYPOINT WAS SET!"
                result = None
                if entrypoint:
                    result, failure = self._execute_safely(entrypoint, item)
                if failure:
                    logger.warning("¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤ CRON FILTER FAILED WITH ¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤")
                    logger.warning(failure)
                    logger.warning("")
                id2, updated_at2 = self.db.bump_batch_item(id, self.db.FAILED if failure else self.db.DONE, failure, updated_at) or (None, None)
                logger.info(f"id2={id2}, updated_at2={updated_at2}, id={id}, updated_at={updated_at}")
            else:
                logger.warning(f"Missing data for item id={id}, type={type}, updated_at={updated_at}")

        # else:
        # logger.info(f"No pending items found")

    def get_now(self):
        return self.db.get_now()

    def get_crons(self):
        return self.db.get_crons(limit=100)

    def log_cron_run(self, cron_id):
        self.db.log_cron_run(cron_id)

    def upsert_cron(self, type="test", data=None, priority=50, source=None):
        """
        Insert a new cron item into the database, ready for execution
        """
        return self.db.insert_cron_item({"priority": priority, "data": data, "type": type, "status": self.db.PENDING, "source": source})

    def delete_all_crons(self, time_sec=30):
        self.db.delete_all()

    def get_status(self):
        status = {"type": self.db.get_type_counts(), "status": self.db.get_status_counts(), "last": self.db.get_last_jobs(limit=10), "type_status": self.db.get_type_status_counts()}
        return status

    def process(self):
        if not self.execute_one_cron_item():
            # Put a status
            # if not self.last_status_time or (datetime.now() - self.last_status_time).total_seconds() > 10.0:
            # 	self.print_status()
            # Lets not get stuck in a loop!
            time.sleep(1)



#!/usr/bin/env python

import os
import pprint
from setuptools import setup, find_namespace_packages

import logging

logger = logging.getLogger(__name__)

setup_requirements = ["pytest-runner", "setuptools_scm"]

version_file = "./VERSION"
readme_file = "./README.md"


def read_file(fname, strip=True):
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    data = ""
    if os.path.exists(fn):
        with open(fn) as f:
            data = f.read()
            data = data.strip() if strip else data
            # logger.info(f"Got data '{data}' from '{fn}'")
    else:
        logger.error(f"Could not find file {fn}")
        logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
    return data


def remove_comment(line, sep="#"):
    i = line.find(sep)
    if i >= 0:
        line = line[:i]
    return line.strip()


def read_requirements_file(fname: str):
    fn = os.path.join(os.path.dirname(os.path.abspath(__file__)), fname)
    print(f"Reading requirements from {fn}")
    lines = []
    if os.path.exists(fn):
        with open(fn) as f:
            for r in f.readlines():
                r = r.strip()
                if len(r) < 1:
                    continue
                r = remove_comment(r)
                if len(r) < 1:
                    continue
                lines.append(r)
    else:
        logger.error(f"Could not find requirements file {fn}")
        logger.warning(f"NOTE: Current working directory is {os.getcwd()}")
    return lines


package = {
    "name": "octomy-clockwork",
    "version": read_file(version_file),
    "author": "Lennart Rolland",
    "author_email": "lennart@octomy.org",
    "maintainer": "Lennart Rolland",
    "maintainer_email": "lennart@octomy.org",
    "description": ("OctoMY scheduler"),
    "license": "GPL-3 LGPL-3 MIT",
    "platforms": ["Linux"],
    "keywords": "octomy clockwork scheduler clock timer task queue",
    "url": "https://gitlab.com/octomy/clockwork",
    "packages": find_namespace_packages(include=['fk.*', ]),
    "scripts": ['bin/octomy-clockwork'],
    "zip_safe": True,
    "long_description": read_file(readme_file),
    "long_description_content_type": "text/markdown",
    "setup_requires": setup_requirements,
    "zip_safe": True,
    "install_requires": read_requirements_file("requirements/requirements.in"),  # Allow flexible deps for install
    "tests_require": read_requirements_file("requirements/test_requirements.txt"),  # Use rigid deps for testing
    "test_suite": "tests",
    "python_requires": "~=3.7.4",
    "data_files": [("clockwork", [version_file])],
    "include_package_data": True,
    # From https://pypi.org/pypi?%3Aaction=list_classifiers
    "classifiers": ["Development Status :: 3 - Alpha", "Intended Audience :: Developers", "Intended Audience :: Other Audience", "Topic :: Utilities", "Natural Language :: English", "Operating System :: POSIX :: Linux", "Programming Language :: Python :: 3.7", "Topic :: Other/Nonlisted Topic"],
}

pprint.pprint(package)
setup(**package)
